@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use Log::Any '$log';
 
-our $VERSION = '0.06'; # VERSION
+our $VERSION = '0.07'; # VERSION
 
 use File::Flock;
 use List::Util qw(max first);
@@ -16,6 +16,7 @@ our @EXPORT_OK = qw(
                        add_group
                        add_user
                        add_user_to_group
+                       add_delete_user_groups
                        delete_group
                        delete_user
                        delete_user_from_group
@@ -1466,6 +1467,7 @@ $SPEC{add_user_to_group} = {
     v => 1.1,
     summary => 'Add user to a group',
     args => {
+        %common_args,
         user => {
             req => 1,
         },
@@ -1500,6 +1502,7 @@ $SPEC{delete_user_from_group} = {
     v => 1.1,
     summary => 'Delete user from a group',
     args => {
+        %common_args,
         user => {
             req => 1,
         },
@@ -1529,6 +1532,81 @@ sub delete_user_from_group {
     );
 }
 
+$SPEC{add_delete_user_groups} = {
+    v => 1.1,
+    summary => 'Add or delete user from one or several groups',
+    description => <<'_',
+
+This can be used to reduce several `add_user_to_group()` and/or
+`delete_user_from_group()` calls to a single call. So:
+
+    add_delete_user_groups(user=>'u',add_to=>['a','b'],delete_from=>['c','d']);
+
+is equivalent to:
+
+    add_user_to_group     (user=>'u', group=>'a');
+    add_user_to_group     (user=>'u', group=>'b');
+    delete_user_from_group(user=>'u', group=>'c');
+    delete_user_from_group(user=>'u', group=>'d');
+
+except that `add_delete_user_groups()` does it in one pass.
+
+_
+    args => {
+        %common_args,
+        user => {
+            req => 1,
+        },
+        add_to => {
+            summary => 'List of group names to add the user as member of',
+            schema => [array => {of=>'str*', default=>[]}],
+        },
+        delete_from => {
+            summary => 'List of group names to remove the user as member of',
+            schema => [array => {of=>'str*', default=>[]}],
+        },
+    },
+};
+sub add_delete_user_groups {
+    my %args = @_;
+    my $user = $args{user} or return [400, "Please specify user"];
+    $user =~ $re_user or return [400, "Invalid user"];
+    my $add  = $args{add_to} // [];
+    my $del  = $args{delete_from} // [];
+
+    # XXX check user exists
+
+    _routine(
+        %args,
+        _lock            => 1,
+        _write_group     => 1,
+        _after_read      => sub {
+            my $stash = shift;
+
+            my $group = $stash->{group};
+            my $changed;
+
+            for my $l (@$group) {
+                my @mm = split /,/, $l->[-1];
+                if ($l->[0] ~~ $add && !($user ~~ @mm)) {
+                    $changed++;
+                    push @mm, $user;
+                }
+                if ($l->[0] ~~ $del && $user ~~ @mm) {
+                    $changed++;
+                    @mm = grep {$_ ne $user} @mm;
+                }
+                if ($changed) {
+                    $l->[-1] = join ",", @mm;
+                }
+            }
+            $stash->{write_group} = 0 unless $changed;
+            $stash->{res} = [200, "OK"];
+            [200];
+        },
+    );
+}
+
 $SPEC{set_user_password} = {
     v => 1.1,
     summary => 'Set user\'s password',
@@ -1688,11 +1766,11 @@ Unix::Passwd::File - Manipulate /etc/{passwd,shadow,group,gshadow} entries
 
 =head1 VERSION
 
-version 0.06
+version 0.07
 
 =head1 SYNOPSIS
 
- use Unix::Passwd::Files;
+ use Unix::Passwd::File;
 
  # list users. by default uses files in /etc (/etc/passwd, /etc/shadow, et al)
  my $res = list_users(); # [200, "OK", ["root", ...]]
@@ -1801,9 +1879,58 @@ L<Setup::Unix::User> and L<Setup::Unix::Group>, which use this module.
 
 L<Rinci>
 
+=head1 DESCRIPTION
+
+
+This module has L<Rinci> metadata.
+
 =head1 FUNCTIONS
 
 
+None are exported by default, but they are exportable.
+
+=head2 add_delete_user_groups(%args) -> [status, msg, result, meta]
+
+Add or delete user from one or several groups.
+
+This can be used to reduce several C<add_user_to_group()> and/or
+C<delete_user_from_group()> calls to a single call. So:
+
+    add_delete_user_groups(user=>'u',add_to=>['a','b'],delete_from=>['c','d']);
+
+is equivalent to:
+
+    add_user_to_group     (user=>'u', group=>'a');
+    add_user_to_group     (user=>'u', group=>'b');
+    delete_user_from_group(user=>'u', group=>'c');
+    delete_user_from_group(user=>'u', group=>'d');
+
+except that C<add_delete_user_groups()> does it in one pass.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<add_to> => I<array> (default: [])
+
+List of group names to add the user as member of.
+
+=item * B<delete_from> => I<array> (default: [])
+
+List of group names to remove the user as member of.
+
+=item * B<etc_dir> => I<str> (default: "/etc")
+
+Specify location of passwd files.
+
+=item * B<user>* => I<any>
+
+=back
+
+Return value:
+
+Returns an enveloped result (an array). First element (status) is an integer containing HTTP status code (200 means OK, 4xx caller error, 5xx function error). Second element (msg) is a string containing error message, or 'OK' if status is 200. Third element (result) is optional, the actual result. Fourth element (meta) is called result metadata and is optional, a hash that contains extra information.
+
 =head2 add_group(%args) -> [status, msg, result, meta]
 
 Add a new group.
@@ -1819,7 +1946,7 @@ Whether to backup when modifying files.
 Backup is written with C<.bak> extension in the same directory. Unmodified file
 will not be backed up. Previous backup will be overwritten.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
@@ -1829,7 +1956,7 @@ Pick a specific new GID.
 
 Adding a new group with duplicate GID is allowed.
 
-=item * B<group> => I<any>
+=item * B<group>* => I<any>
 
 =item * B<max_gid> => I<int> (default: 65535)
 
@@ -1870,11 +1997,11 @@ Whether to backup when modifying files.
 Backup is written with C<.bak> extension in the same directory. Unmodified file
 will not be backed up. Previous backup will be overwritten.
 
-=item * B<encpass>* => I<str>
+=item * B<encpass> => I<str>
 
 Encrypted password.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
@@ -1882,7 +2009,7 @@ Specify location of passwd files.
 
 The date of expiration of the account, expressed as the number of days since Jan 1, 1970.
 
-=item * B<gecos>* => I<str>
+=item * B<gecos> => I<str>
 
 Usually, it contains the full username.
 
@@ -1901,13 +2028,13 @@ will be created if does not already exist. You can pick another group here,
 which must already exist (and in this case, the group with the same name as user
 will not be created).
 
-=item * B<home>* => I<str>
+=item * B<home> => I<str>
 
 User's home directory.
 
 =item * B<last_pwchange> => I<int>
 
-The date of the last password change, expressed as the number of days since Jan 1, 1970..
+The date of the last password change, expressed as the number of days since Jan 1, 1970.
 
 =item * B<max_gid> => I<any>
 
@@ -1939,7 +2066,7 @@ Pick a range for new UID.
 If a free UID between C<min_uid> and C<max_uid> is not found, error 412 is
 returned.
 
-=item * B<pass>* => I<str>
+=item * B<pass> => I<str>
 
 Password, generally should be "x" which means password is encrypted in shadow.
 
@@ -1951,7 +2078,7 @@ The number of days after a password has expired (see max_pass_age) during which
 
 The number of days before a password is going to expire (see max_pass_age) during which the user should be warned.
 
-=item * B<shell>* => I<str>
+=item * B<shell> => I<str>
 
 User's home directory.
 
@@ -1961,7 +2088,7 @@ Pick a specific new UID.
 
 Adding a new user with duplicate UID is allowed.
 
-=item * B<user> => I<any>
+=item * B<user>* => I<any>
 
 =back
 
@@ -1977,9 +2104,13 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<group> => I<any>
+=item * B<etc_dir> => I<str> (default: "/etc")
 
-=item * B<user> => I<any>
+Specify location of passwd files.
+
+=item * B<group>* => I<any>
+
+=item * B<user>* => I<any>
 
 =back
 
@@ -2002,11 +2133,11 @@ Whether to backup when modifying files.
 Backup is written with C<.bak> extension in the same directory. Unmodified file
 will not be backed up. Previous backup will be overwritten.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<group> => I<any>
+=item * B<group>* => I<any>
 
 =back
 
@@ -2029,11 +2160,11 @@ Whether to backup when modifying files.
 Backup is written with C<.bak> extension in the same directory. Unmodified file
 will not be backed up. Previous backup will be overwritten.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<user> => I<any>
+=item * B<user>* => I<any>
 
 =back
 
@@ -2049,9 +2180,13 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<group> => I<any>
+=item * B<etc_dir> => I<str> (default: "/etc")
 
-=item * B<user> => I<any>
+Specify location of passwd files.
+
+=item * B<group>* => I<any>
+
+=item * B<user>* => I<any>
 
 =back
 
@@ -2071,13 +2206,13 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<gid>* => I<int>
+=item * B<gid> => I<int>
 
-=item * B<group>* => I<str>
+=item * B<group> => I<str>
 
 =item * B<with_field_names> => I<bool> (default: 1)
 
@@ -2101,7 +2236,7 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
@@ -2119,7 +2254,7 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
@@ -2141,13 +2276,13 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<uid>* => I<int>
+=item * B<uid> => I<int>
 
-=item * B<user>* => I<str>
+=item * B<user> => I<str>
 
 =item * B<with_field_names> => I<bool> (default: 1)
 
@@ -2175,11 +2310,11 @@ Arguments ('*' denotes required arguments):
 
 If true, return all fields instead of just group names.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<user> => I<any>
+=item * B<user>* => I<any>
 
 =item * B<with_field_names> => I<bool> (default: 1)
 
@@ -2196,7 +2331,7 @@ Return value:
 
 Returns an enveloped result (an array). First element (status) is an integer containing HTTP status code (200 means OK, 4xx caller error, 5xx function error). Second element (msg) is a string containing error message, or 'OK' if status is 200. Third element (result) is optional, the actual result. Fourth element (meta) is called result metadata and is optional, a hash that contains extra information.
 
-=head2 group_exists(%args) -> [status, msg, result, meta]
+=head2 group_exists(%args) -> bool
 
 Check whether group exists.
 
@@ -2204,21 +2339,19 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<gid>* => I<int>
+=item * B<gid> => I<int>
 
-=item * B<group>* => I<str>
+=item * B<group> => I<str>
 
 =back
 
 Return value:
 
-Returns an enveloped result (an array). First element (status) is an integer containing HTTP status code (200 means OK, 4xx caller error, 5xx function error). Second element (msg) is a string containing error message, or 'OK' if status is 200. Third element (result) is optional, the actual result. Fourth element (meta) is called result metadata and is optional, a hash that contains extra information.
-
-=head2 is_member(%args) -> [status, msg, result, meta]
+=head2 is_member(%args) -> bool
 
 Check whether user is member of a group.
 
@@ -2226,20 +2359,18 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<group> => I<any>
+=item * B<group>* => I<any>
 
-=item * B<user> => I<any>
+=item * B<user>* => I<any>
 
 =back
 
 Return value:
 
-Returns an enveloped result (an array). First element (status) is an integer containing HTTP status code (200 means OK, 4xx caller error, 5xx function error). Second element (msg) is a string containing error message, or 'OK' if status is 200. Third element (result) is optional, the actual result. Fourth element (meta) is called result metadata and is optional, a hash that contains extra information.
-
 =head2 list_groups(%args) -> [status, msg, result, meta]
 
 List Unix groups in group file.
@@ -2252,7 +2383,7 @@ Arguments ('*' denotes required arguments):
 
 If true, return all fields instead of just group names.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
@@ -2283,7 +2414,7 @@ Arguments ('*' denotes required arguments):
 
 If true, return all fields instead of just usernames.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
@@ -2318,7 +2449,7 @@ Arguments ('*' denotes required arguments):
 
 If true, return all fields instead of just names.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
@@ -2343,7 +2474,7 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<admins>* => I<str>
+=item * B<admins> => I<str>
 
 It must be a comma-separated list of user names, or empty.
 
@@ -2354,27 +2485,27 @@ Whether to backup when modifying files.
 Backup is written with C<.bak> extension in the same directory. Unmodified file
 will not be backed up. Previous backup will be overwritten.
 
-=item * B<encpass>* => I<str>
+=item * B<encpass> => I<str>
 
 Encrypted password.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<gid>* => I<int>
+=item * B<gid> => I<int>
 
 Numeric group ID.
 
-=item * B<group>* => I<str>
+=item * B<group> => I<str>
 
 Group name.
 
-=item * B<members>* => I<str>
+=item * B<members> => I<str>
 
 List of usernames that are members of this group, separated by commas.
 
-=item * B<pass>* => I<str>
+=item * B<pass> => I<str>
 
 Password, generally should be "x" which means password is encrypted in gshadow.
 
@@ -2402,11 +2533,11 @@ Whether to backup when modifying files.
 Backup is written with C<.bak> extension in the same directory. Unmodified file
 will not be backed up. Previous backup will be overwritten.
 
-=item * B<encpass>* => I<str>
+=item * B<encpass> => I<str>
 
 Encrypted password.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
@@ -2414,21 +2545,21 @@ Specify location of passwd files.
 
 The date of expiration of the account, expressed as the number of days since Jan 1, 1970.
 
-=item * B<gecos>* => I<str>
+=item * B<gecos> => I<str>
 
 Usually, it contains the full username.
 
-=item * B<gid>* => I<int>
+=item * B<gid> => I<int>
 
 Numeric primary group ID for this user.
 
-=item * B<home>* => I<str>
+=item * B<home> => I<str>
 
 User's home directory.
 
 =item * B<last_pwchange> => I<int>
 
-The date of the last password change, expressed as the number of days since Jan 1, 1970..
+The date of the last password change, expressed as the number of days since Jan 1, 1970.
 
 =item * B<max_pass_age> => I<int>
 
@@ -2438,7 +2569,7 @@ The number of days after which the user will have to change her password.
 
 The number of days the user will have to wait before she will be allowed to change her password again.
 
-=item * B<pass>* => I<str>
+=item * B<pass> => I<str>
 
 Password, generally should be "x" which means password is encrypted in shadow.
 
@@ -2450,15 +2581,15 @@ The number of days after a password has expired (see max_pass_age) during which
 
 The number of days before a password is going to expire (see max_pass_age) during which the user should be warned.
 
-=item * B<shell>* => I<str>
+=item * B<shell> => I<str>
 
 User's home directory.
 
-=item * B<uid>* => I<int>
+=item * B<uid> => I<int>
 
 Numeric user ID.
 
-=item * B<user>* => I<str>
+=item * B<user> => I<str>
 
 User (login) name.
 
@@ -2483,13 +2614,13 @@ Whether to backup when modifying files.
 Backup is written with C<.bak> extension in the same directory. Unmodified file
 will not be backed up. Previous backup will be overwritten.
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<pass> => I<any>
+=item * B<pass>* => I<any>
 
-=item * B<user> => I<any>
+=item * B<user>* => I<any>
 
 =back
 
@@ -2497,7 +2628,7 @@ Return value:
 
 Returns an enveloped result (an array). First element (status) is an integer containing HTTP status code (200 means OK, 4xx caller error, 5xx function error). Second element (msg) is a string containing error message, or 'OK' if status is 200. Third element (result) is optional, the actual result. Fourth element (meta) is called result metadata and is optional, a hash that contains extra information.
 
-=head2 user_exists(%args) -> [status, msg, result, meta]
+=head2 user_exists(%args) -> bool
 
 Check whether user exists.
 
@@ -2505,20 +2636,18 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<etc_dir>* => I<str> (default: "/etc")
+=item * B<etc_dir> => I<str> (default: "/etc")
 
 Specify location of passwd files.
 
-=item * B<uid>* => I<int>
+=item * B<uid> => I<int>
 
-=item * B<user>* => I<str>
+=item * B<user> => I<str>
 
 =back
 
 Return value:
 
-Returns an enveloped result (an array). First element (status) is an integer containing HTTP status code (200 means OK, 4xx caller error, 5xx function error). Second element (msg) is a string containing error message, or 'OK' if status is 200. Third element (result) is optional, the actual result. Fourth element (meta) is called result metadata and is optional, a hash that contains extra information.
-
 =head1 AUTHOR
 
 Steven Haryanto <stevenharyanto@gmail.com>
